#!/usr/bin/env python3

import os
import platform
import sys
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Union

import click

from .layout import KeyboardLayout, load_layout
from .xkb_manager import WAYLAND, Index, XKBManager


@click.group()
def cli() -> None:
    if platform.system() != "Linux":
        sys.exit("This command is only compatible with GNU/Linux, sorry.")


@cli.command()
@click.argument(
    "filepath", type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
@click.option(
    "--angle-mod/--no-angle-mod",
    default=False,
    help="Apply Angle-Mod (which is a [ZXCVB] permutation with the LSGT key (a.k.a. ISO key))",
)
def apply(filepath: Path, angle_mod: bool) -> None:
    """Apply a Kalamine layout."""

    if WAYLAND:
        sys.exit(
            "You appear to be running Wayland, which does not support this operation."
        )

    layout = KeyboardLayout(load_layout(filepath), angle_mod)
    with tempfile.NamedTemporaryFile(
        mode="w+", suffix=".xkb_keymap", encoding="utf-8"
    ) as temp_file:
        temp_file.write(layout.xkb_keymap)
        os.system(f"xkbcomp -w0 {temp_file.name} $DISPLAY")


@cli.command()
@click.argument(
    "layouts", nargs=-1, type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
@click.option(
    "--angle-mod/--no-angle-mod",
    default=False,
    help="Apply Angle-Mod (which is a [ZXCVB] permutation with the LSGT key (a.k.a. ISO key))",
)
def install(layouts: List[Path], angle_mod: bool) -> None:
    """Install a list of Kalamine layouts."""

    if not layouts:
        return

    kb_locales = set()
    kb_layouts = []
    for file in layouts:
        layout_file = load_layout(file)
        layout = KeyboardLayout(layout_file, angle_mod)
        kb_layouts.append(layout)
        kb_locales.add(layout.meta["locale"])

    def xkb_install(xkb: XKBManager) -> Index:
        for layout in kb_layouts:
            xkb.add(layout)
        index = xkb.index  # gets erased with xkb.update()
        xkb.clean()
        xkb.update()
        print()
        print("Successfully installed.")
        return dict(index)

    # EAFP (Easier to Ask Forgiveness than Permission)
    try:
        xkb_root = XKBManager(root=True)
        xkb_index = xkb_install(xkb_root)
        print(f"On XOrg, you can try the layout{'s' if len(layouts) > 1 else ''} with:")
        for locale, variants in xkb_index.items():
            for name in variants.keys():
                print(f"    setxkbmap {locale} -variant {name}")
        print()

    except PermissionError:
        print(xkb_root.path)
        print("    Not writable: switching to user-space.")
        print()
        if not WAYLAND:
            print(
                "You appear to be running XOrg. You need sudo privileges to install keyboard layouts:"
            )
            for filepath in layouts:
                print(f'    sudo env "PATH=$PATH" xkalamine install {filepath}')
            sys.exit(1)

        xkb_home = XKBManager()
        xkb_home.ensure_xkb_config_is_ready()
        xkb_install(xkb_home)
        print("Warning: user-space layouts only work with Wayland.")
        print()


@cli.command()
@click.argument("mask")  # [locale]/[name]
def remove(mask: str) -> None:
    """Remove a list of Kalamine layouts."""

    def xkb_remove(root: bool = False) -> None:
        xkb = XKBManager(root=root)
        xkb.clean()
        for locale, variants in xkb.list(mask).items():
            for name in variants.keys():
                xkb.remove(locale, name)
        xkb.update()

    # EAFP (Easier to Ask Forgiveness than Permission)
    try:
        xkb_remove(root=True)
    except PermissionError:
        if not WAYLAND:
            print(
                "You appear to be running XOrg. You need sudo privileges to remove keyboard layouts:"
            )
            print(f'    sudo env "PATH=$PATH" xkalamine remove {mask}')
            sys.exit(1)
        xkb_remove()


@cli.command(name="list")
@click.option("-a", "--all", "all_flag", is_flag=True)
@click.argument("mask", default="*")
def list_command(mask: str, all_flag: bool) -> None:
    """List installed Kalamine layouts."""

    for root in [True, False]:
        # XXX this very weird type means we've done something silly here
        filtered: Dict[str, Union[Optional[KeyboardLayout], str]] = {}
        xkb = XKBManager(root=root)
        layouts = xkb.list_all(mask) if all_flag else xkb.list(mask)
        for locale, variants in sorted(layouts.items()):
            for name, desc in sorted(variants.items()):
                filtered[f"{locale}/{name}"] = desc

        if mask == "*" and root and xkb.has_custom_symbols():
            filtered["custom"] = ""

        if filtered:
            print(xkb.path)
            for key, value in filtered.items():
                print(f"    {key:<24} {value}")


if __name__ == "__main__":
    cli()



import ctypes
import os
import subprocess
import sys
from pathlib import Path
from shutil import move, rmtree
from stat import S_IREAD, S_IWUSR

import tomli

from .help import create_layout_content
from .layout import KeyboardLayout


class MsklcManager:
    def __init__(
        self,
        layout: "KeyboardLayout",
        msklc_dir: Path,
        working_dir: Path = Path(os.getcwd()),
        verbose: bool = False,
    ) -> None:
        self._layout = layout
        self._msklc_dir = msklc_dir
        self._verbose = verbose
        self._working_dir = working_dir

    # This need the klc file to be created first
    def create_c_files(self):
        """Call kbdutools to generate C files"""
        kbdutools = self._msklc_dir / Path("bin/i386/kbdutool.exe")
        cur = os.getcwd()
        os.chdir(self._working_dir)
        ret = subprocess.run(
            [kbdutools, "-u", "-s", self._layout.meta["name8"] + ".klc"],
            capture_output=not self._verbose,
        )
        os.chdir(cur)
        ret.check_returncode()

    # Check if the driver is already installed
    # as this will cause MSKLC to launch the GUI instead of creating the installer
    def _is_already_installed(self) -> bool:
        sys32 = Path(os.environ["WINDIR"]) / Path("System32")
        dll = sys32 / Path(f'{self._layout.meta["name8"]}.dll')
        return dll.exists()

    def build_msklc_installer(
        self,
    ) -> bool:
        name8 = self._layout.meta["name8"]

        if (self._working_dir / Path(name8)).exists():
            print(
                f"WARN: `{self._working_dir / Path(name8)}` already exists, "
                "assuming installer sits there."
            )
            return True

        if self._is_already_installed():
            print(
                "Error: layout already installed and "
                "installer package not found in the current directory.\n"
                "Either uninstall the layout manually, or put the installer "
                f"folder in the current directory: ({self._working_dir})"
            )
            return False

        # Create a dummy klc file to generate the installer.
        # The file must have a correct name to be reflected in the installer.
        dummy_toml = create_layout_content(
            self._layout.geometry, self._layout.has_altgr, self._layout.has_1dk
        )
        dummy_toml = dummy_toml.replace(
            "custom QWERTY layout", self._layout.meta["description"]
        )
        dummy_toml = dummy_toml.replace("qwerty-custom", self._layout.meta["name"])
        dummy_toml = dummy_toml.replace("custom", self._layout.meta["name8"])
        dummy_toml = dummy_toml.replace("nobody", self._layout.meta["author"])
        dummy_toml = dummy_toml.replace(
            "https://OneDeadKey.github.com/kalamine", self._layout.meta["url"]
        )
        dummy_toml = dummy_toml.replace("0.0.1", self._layout.meta["version"])
        dummy_toml = dummy_toml.replace('"us"', f'"{self._layout.meta["locale"]}"')

        dummy_layout = KeyboardLayout(tomli.loads(dummy_toml))
        dummy_klc = dummy_layout.klc
        klc_file = Path(self._working_dir) / Path(f'{self._layout.meta["name8"]}.klc')
        with klc_file.open("w", encoding="utf-16le", newline="\n") as file:
            file.write(dummy_klc)
        msklc = self._msklc_dir / Path("MSKLC.exe")
        subprocess.run([msklc, klc_file, "-build"], capture_output=not self._verbose)

        # move the installer from "My Documents" to current dir
        if sys.platform == "win32":  # let mypy know this is win32-specific
            CSIDL_PERSONAL = 5  # My Documents
            SHGFP_TYPE_CURRENT = 0  # Get current, not default value
            buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
            ctypes.windll.shell32.SHGetFolderPathW(
                None, CSIDL_PERSONAL, None, SHGFP_TYPE_CURRENT, buf
            )
            my_docs = Path(buf.value)
            installer = my_docs / Path(name8)

            if installer.exists():
                move(str(installer), str(self._working_dir / Path(name8)))

        return True

    def build_msklc_dll(
        self,
    ):
        name8 = self._layout.meta["name8"]
        prev = os.getcwd()
        os.chdir(self._working_dir)
        INST_DIR = self._working_dir / Path(name8)
        dll_dirs = ["i386", "amd64", "ia64", "wow64"]
        for dll_dir in dll_dirs:
            full_dir = INST_DIR / Path(dll_dir)
            if not full_dir.exists():
                raise Exception(f"{full_dir} doesn't exist")
            else:
                rmtree(full_dir)
                os.mkdir(full_dir)

        klc_file = self._working_dir / Path(f"{name8}.klc")

        # create correct klc
        with klc_file.open("w", encoding="utf-16le", newline="\n") as file:
            file.write(self._layout.klc)

        self.create_c_files()
        c_file = klc_file.with_suffix(".RC")
        with c_file.open("w", encoding="utf-16le", newline="\n") as file:
            file.write(self._layout.klc_rc)
        c_file = klc_file.with_suffix(".C")
        with c_file.open("w", encoding="utf-16le", newline="\n") as file:
            file.write(self._layout.klc_c)
        c_files = [".C", ".RC", ".H", ".DEF"]

        # Make files read-only to prevent MSKLC from overwriting them.
        for suffix in c_files:
            os.chmod(klc_file.with_suffix(suffix), S_IREAD)

        # build correct DLLs

        kbdutools = self._msklc_dir / Path("bin/i386/kbdutool.exe")
        dll = klc_file.with_suffix(".dll")

        for arch_flag, arch in [
            ("-x", "i386"),
            ("-m", "amd64"),
            ("-i", "ia64"),
            ("-o", "wow64"),
        ]:
            subprocess.run(
                [kbdutools, "-u", arch_flag, klc_file], capture_output=not self._verbose
            ).check_returncode()
            move(dll, INST_DIR / Path(arch))

        # Restore write permission
        for suffix in c_files:
            os.chmod(klc_file.with_suffix(suffix), S_IWUSR)
        os.chdir(prev)



import json
from typing import TYPE_CHECKING, Dict, List, Tuple

from .utils import DEAD_KEYS, LAYER_KEYS, ODK_ID, Layer, load_data

if TYPE_CHECKING:
    from .layout import KeyboardLayout

DK_INDEX = {}
for dk in DEAD_KEYS:
    DK_INDEX[dk.char] = dk


###
# Helpers
#

SCAN_CODES = load_data("scan_codes")
XKB_KEY_SYM = load_data("key_sym")


def hex_ord(char: str) -> str:
    return hex(ord(char))[2:].zfill(4)


def xml_proof(char: str) -> str:
    if char not in '<&"\u0020\u00a0\u202f>':
        return char
    return f"&#x{hex_ord(char)};"


def xml_proof_id(symbol: str) -> str:
    return symbol[2:-1] if symbol.startswith("&#x") else symbol


###
# GNU/Linux: XKB
# - standalone xkb keymap file to be used by `xkbcomp` (XOrg only)
# - xkb symbols/patch for XOrg (system-wide) & Wayland (system-wide/user-space)
#


def xkb_keymap(layout: "KeyboardLayout", xkbcomp: bool = False) -> List[str]:
    """Linux layout."""

    show_description = True
    eight_level = layout.has_altgr and layout.has_1dk and not xkbcomp
    odk_symbol = "ISO_Level5_Latch" if eight_level else "ISO_Level3_Latch"
    max_length = 16  # `ISO_Level3_Latch` should be the longest symbol name

    output: List[str] = []
    for key_name in LAYER_KEYS:
        if key_name.startswith("-"):  # separator
            if output:
                output.append("")
            output.append("//" + key_name[1:])
            continue

        descs = []
        symbols = []
        for layer in layout.layers.values():
            if key_name in layer:
                keysym = layer[key_name]
                desc = keysym
                # dead key?
                if keysym in DK_INDEX:
                    name = DK_INDEX[keysym].name
                    desc = layout.dead_keys[keysym][keysym]
                    symbol = odk_symbol if keysym == ODK_ID else f"dead_{name}"
                # regular key: use a keysym if possible, utf-8 otherwise
                elif keysym in XKB_KEY_SYM and len(XKB_KEY_SYM[keysym]) <= max_length:
                    symbol = XKB_KEY_SYM[keysym]
                else:
                    symbol = f"U{hex_ord(keysym).upper()}"
            else:
                desc = " "
                symbol = "VoidSymbol"

            descs.append(desc)
            symbols.append(symbol.ljust(max_length))

        key = "{{[ {0}, {1}, {2}, {3}]}}"  # 4-level layout by default
        description = "{0} {1} {2} {3}"
        if layout.has_altgr and layout.has_1dk:
            # 6 layers are needed: they won't fit on the 4-level format.
            if xkbcomp:  # user-space XKB keymap file (standalone)
                # standalone XKB files work best with a dual-group solution:
                # one 4-level group for base+1dk, one two-level group for AltGr
                key = "{{[ {}, {}, {}, {}],[ {}, {}]}}"
                description = "{} {} {} {} {} {}"
            else:  # eight_level XKB symbols (Neo-like)
                key = "{{[ {0}, {1}, {4}, {5}, {2}, {3}]}}"
                description = "{0} {1} {4} {5} {2} {3}"
        elif layout.has_altgr:
            del symbols[3]
            del symbols[2]
            del descs[3]
            del descs[2]

        line = f"key <{key_name.upper()}> {key.format(*symbols)};"
        if show_description:
            line += (" // " + description.format(*descs)).rstrip()
            if line.endswith("\\"):
                line += " "  # escape trailing backslash
        output.append(line)

    return output


###
# Windows: AHK
# To be used by AutoHotKey v1.1: https://autohotkey.com
# During our tests, AHK 2.0 has raised serious performance and stability issues.
# FWIW, PKL and EPKL still rely on AHK 1.1, too.


def ahk_keymap(layout: "KeyboardLayout", altgr: bool = False) -> List[str]:
    """AHK layout, main and AltGr layers."""

    prefixes = [" ", "+", "", "", " <^>!", "<^>!+"]
    specials = " \u00a0\u202f‘’'\"^`~"
    esc_all = True  # set to False to ease the debug (more readable AHK script)

    def ahk_escape(key: str) -> str:
        if len(key) == 1:
            return f"U+{ord(key):04x}" if (esc_all or key in specials) else key
        return f"{key}`" if key.endswith("`") else key  # deadkey identifier

    def ahk_actions(symbol: str) -> Dict[str, str]:
        actions = {}
        for key, dk in layout.dead_keys.items():
            dk_id = ahk_escape(key)
            if symbol == "spce":
                actions[dk_id] = ahk_escape(dk[" "])
            elif symbol in dk:
                actions[dk_id] = ahk_escape(dk[symbol])
        return actions

    output = []
    for key_name in LAYER_KEYS:
        if key_name.startswith("-"):
            output.append(f"; {key_name[1:]}")
            output.append("")
            continue

        if key_name in ["ae13", "ab11"]:  # ABNT / JIS keys
            continue  # these two keys are not supported yet

        sc = f"SC{SCAN_CODES['klc'][key_name]}"
        for i in (
            [Layer.ALTGR, Layer.ALTGR_SHIFT] if altgr else [Layer.BASE, Layer.SHIFT]
        ):
            layer = layout.layers[i]
            if key_name not in layer:
                continue

            symbol = layer[key_name]
            sym = ahk_escape(symbol)

            if symbol in layout.dead_keys:
                actions = {sym: layout.dead_keys[symbol][symbol]}
            elif key_name == "spce":
                actions = ahk_actions(key_name)
            else:
                actions = ahk_actions(symbol)

            desc = f" ; {symbol}" if symbol != sym else ""
            act = json.dumps(actions, ensure_ascii=False)
            output.append(f'{prefixes[i]}{sc}::SendKey("{sym}", {act}){desc}')

        if output[-1]:
            output.append("")

    return output


def ahk_shortcuts(layout: "KeyboardLayout") -> List[str]:
    """AHK layout, shortcuts."""

    prefixes = [" ^", "^+"]
    enabled = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

    output = []
    for key_name in LAYER_KEYS:
        if key_name.startswith("-"):
            output.append(f"; {key_name[1:]}")
            output.append("")
            continue

        if key_name in ["ae13", "ab11"]:  # ABNT / JIS keys
            continue  # these two keys are not supported yet

        sc = f"SC{SCAN_CODES['klc'][key_name]}"
        for i in [Layer.BASE, Layer.SHIFT]:
            layer = layout.layers[i]
            if key_name not in layer:
                continue

            symbol = layer[key_name]
            if symbol in enabled:
                output.append(f"{prefixes[i]}{sc}::Send {prefixes[i]}{symbol}")

        if output[-1]:
            output.append("")

    return output


###
# Windows: KLC
# To be used by the MS Keyboard Layout Creator to generate an installer.
# https://www.microsoft.com/en-us/download/details.aspx?id=102134
# https://levicki.net/articles/2006/09/29/HOWTO_Build_keyboard_layouts_for_Windows_x64.php
# Also supported by KbdEdit: http://www.kbdedit.com/ (non-free).
# Note: blank lines and comments in KLC sections are removed from the output
# file because they are not recognized by KBDEdit (as of v19.8.0).
#


# return the corresponding char for a symbol
def get_chr(symbol: str) -> str:
    if len(symbol) > 1 and symbol.endswith("@"):
        # remove dead key symbol for dict access
        key = symbol[:-1]
    else:
        key = symbol

    if len(symbol) == 4:
        char = chr(int(key, base=16))
    else:
        char = symbol

    return char


oem_idx = 0


def klc_virtual_key(layout: "KeyboardLayout", symbols: list, scan_code: str) -> str:
    if (layout.meta["geometry"] == "ISO" and scan_code == "56") or symbols[0] == "-1":
        # manage the ISO key (between shift and Z on ISO keyboards).
        # We're assuming that its scancode is always 56
        # https://www.win.tue.nl/~aeb/linux/kbd/scancodes.html
        return "OEM_102"

    base = get_chr(symbols[0])
    shifted = get_chr(symbols[1])

    # Can’t use `isdigit()` because `²` is a digit but we don't want that as a VK
    allowed_digit = "0123456789"
    # We assume that digit row always have digit as VK
    if base in allowed_digit:
        return base
    elif shifted in allowed_digit:
        return shifted

    if shifted.isascii() and shifted.isalpha():
        return shifted

    # VK_OEM_* case
    if base == "," or shifted == ",":
        return "OEM_COMMA"
    elif base == "." or shifted == ".":
        return "OEM_PERIOD"
    elif base == "+" or shifted == "+":
        return "OEM_PLUS"
    elif base == "-" or shifted == "-":
        return "OEM_MINUS"
    elif base == " ":
        return "SPACE"
    else:
        MAX_OEM = 8
        # We affect abitrary OEM VK and it will not match the one
        # in distributed layout. It can cause issue if a application
        # is awaiting a particular OEM_ for a hotkey
        global oem_idx
        oem_idx += 1
        if oem_idx <= MAX_OEM:
            return "OEM_" + str(oem_idx)
        else:
            raise Exception("Too many OEM keys")


def klc_keymap(layout: "KeyboardLayout") -> List[str]:
    """Windows layout, main part."""

    supported_symbols = "1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

    global oem_idx
    oem_idx = 0  # Python trick to do equivalent of C static variable
    output = []
    for key_name in LAYER_KEYS:
        if key_name.startswith("-"):
            continue

        if key_name in ["ae13", "ab11"]:  # ABNT / JIS keys
            continue  # these two keys are not supported yet

        symbols = []
        description = "//"
        is_alpha = False

        for i in [Layer.BASE, Layer.SHIFT, Layer.ALTGR, Layer.ALTGR_SHIFT]:
            layer = layout.layers[i]

            if key_name in layer:
                symbol = layer[key_name]
                desc = symbol
                if symbol in layout.dead_keys:
                    desc = layout.dead_keys[symbol][" "]
                    symbol = hex_ord(desc) + "@"
                else:
                    if i == Layer.BASE:
                        is_alpha = symbol.upper() != symbol
                    if symbol not in supported_symbols:
                        symbol = hex_ord(symbol)
                symbols.append(symbol)
            else:
                desc = " "
                symbols.append("-1")
            description += " " + desc

        scan_code = SCAN_CODES["klc"][key_name]

        virtual_key = klc_virtual_key(layout, symbols, scan_code)

        if layout.has_altgr:
            output.append(
                "\t".join(
                    [
                        scan_code,
                        virtual_key,
                        "1" if is_alpha else "0",  # affected by CapsLock?
                        symbols[0],
                        symbols[1],  # base layer
                        "-1",
                        "-1",  # ctrl layer
                        symbols[2],
                        symbols[3],  # altgr layer
                        description.strip(),
                    ]
                )
            )
        else:
            output.append(
                "\t".join(
                    [
                        scan_code,
                        virtual_key,
                        "1" if is_alpha else "0",  # affected by CapsLock?
                        symbols[0],
                        symbols[1],  # base layer
                        "-1",
                        "-1",  # ctrl layer
                        description.strip(),
                    ]
                )
            )

    return output


def klc_deadkeys(layout: "KeyboardLayout") -> List[str]:
    """Windows layout, dead keys."""

    output = []

    for k in DK_INDEX:
        if k not in layout.dead_keys:
            continue
        dk = layout.dead_keys[k]

        output.append(f"// DEADKEY: {DK_INDEX[k].name.upper()} //" + "{{{")
        output.append(f"DEADKEY\t{hex_ord(dk[' '])}")

        for base, alt in dk.items():
            if base == k and alt in base:
                continue

            if base in layout.dead_keys:
                base = layout.dead_keys[base][" "]

            if alt in layout.dead_keys:
                alt = layout.dead_keys[alt][" "]
                ext = hex_ord(alt) + "@"
            else:
                ext = hex_ord(alt)

            output.append(f"{hex_ord(base)}\t{ext}\t// {base} -> {alt}")

        output.append("//}}}")
        output.append("")

    return output[:-1]


def klc_dk_index(layout: "KeyboardLayout") -> List[str]:
    """Windows layout, dead key index."""

    output = []
    for k in DK_INDEX:
        if k not in layout.dead_keys:
            continue
        dk = layout.dead_keys[k]
        output.append(f"{hex_ord(dk[' '])}\t\"{DK_INDEX[k].name.upper()}\"")
    return output


def c_keymap(layout: "KeyboardLayout") -> List[str]:
    """Windows C layout, main part."""

    supported_symbols = "1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

    global oem_idx
    oem_idx = 0  # Python trick to do equivalent of C static variable
    output = []
    for key_name in LAYER_KEYS:
        if key_name.startswith("-"):
            continue

        if key_name in ["ae13", "ab11"]:  # ABNT / JIS keys
            continue  # these two keys are not supported yet

        symbols = []
        dead_symbols = []
        is_alpha = False
        has_dead_key = False

        for i in [Layer.BASE, Layer.SHIFT, Layer.ALTGR, Layer.ALTGR_SHIFT]:
            layer = layout.layers[i]

            if key_name in layer:
                symbol = layer[key_name]
                desc = symbol
                dead = "WCH_NONE"
                if symbol in layout.dead_keys:
                    desc = layout.dead_keys[symbol][" "]
                    symbol = "WCH_DEAD"
                    dead = hex_ord(desc)
                    has_dead_key = True
                else:
                    if i == Layer.BASE:
                        is_alpha = symbol.upper() != symbol
                    if symbol not in supported_symbols:
                        symbol = hex_ord(symbol)
                symbols.append(symbol)
                dead_symbols.append(dead)
            else:
                desc = " "
                symbols.append("WCH_NONE")
                dead_symbols.append("WCH_NONE")

        scan_code = SCAN_CODES["klc"][key_name]
        virtual_key = klc_virtual_key(layout, symbols, scan_code)
        if len(virtual_key) == 1:
            virtual_key_id = f"'{virtual_key}'"
        else:
            virtual_key_id = f"VK_{virtual_key}"

        def process_symbol(symbol: str) -> str:
            if len(symbol) == 4:
                return f"0x{symbol}"
            if len(symbol) == 1:
                return f"'{symbol}'"
            return symbol

        symbols[:] = [process_symbol(symbol) for symbol in symbols]
        dead_symbols[:] = [process_symbol(symbol) for symbol in dead_symbols]

        def key_list(
            key_syms: List[str],
            virt_key: str = "0xff",
            has_altgr: bool = False,
            is_alpha: bool = False,
        ) -> str:
            cols = [
                virt_key,
                "CAPLOK" if is_alpha else "0",  # affected by CapsLock?
                key_syms[0],
                key_syms[1],  # base layer
                "WCH_NONE",
                "WCH_NONE",  # ctrl layer
            ]
            if has_altgr:
                cols += [
                    key_syms[2],
                    key_syms[3],
                ]
            return "\t,".join(cols)

        output.append(
            f"\t{{{key_list(symbols, virtual_key_id, layout.has_altgr, is_alpha)}}},"
        )
        if has_dead_key:
            output.append(
                f"\t{{{key_list(dead_symbols, has_altgr=layout.has_altgr, is_alpha=is_alpha)}}},"
            )

    return output


def c_deadkeys(layout: "KeyboardLayout") -> List[str]:
    """Windows C layout, dead keys."""

    output = []

    for k in DK_INDEX:
        if k not in layout.dead_keys:
            continue
        dk = layout.dead_keys[k]

        output.append(f"// DEADKEY: {DK_INDEX[k].name.upper()}")

        for base, alt in dk.items():
            if base == k and alt in base:
                continue

            if base in layout.dead_keys:
                base = layout.dead_keys[base][" "]

            if alt in layout.dead_keys:
                alt = layout.dead_keys[alt][" "]
                dead_alt = "0x0001"
            else:
                dead_alt = "0x0000"
            ext = hex_ord(alt)

            output.append(
                f"DEADTRANS(0x{hex_ord(base)}\t, 0x{hex_ord(dk[' '])}\t, 0x{ext}\t, {dead_alt}), // {base} -> {alt}"
            )

        output.append("")

    return output[:-1]


def c_dk_index(layout: "KeyboardLayout") -> List[str]:
    """Windows layout, dead key index."""

    output = []
    for k in DK_INDEX:
        if k not in layout.dead_keys:
            continue
        term = layout.dead_keys[k][" "]
        output.append(f'L"\\\\x{hex_ord(term)}"\tL"{DK_INDEX[k].name.upper()}",')
    return output


###
# macOS: keylayout
# https://developer.apple.com/library/content/technotes/tn2056/
#


def osx_keymap(layout: "KeyboardLayout") -> List[List[str]]:
    """macOS layout, main part."""

    ret_str = []
    for index in range(5):
        layer = layout.layers[
            [Layer.BASE, Layer.SHIFT, Layer.BASE, Layer.ALTGR, Layer.ALTGR_SHIFT][index]
        ]
        caps = index == 2

        def has_dead_keys(letter: str) -> bool:
            if letter in "\u0020\u00a0\u202f":  # space
                return True
            for k in layout.dead_keys:
                if letter in layout.dead_keys[k]:
                    return True
            return False

        output: List[str] = []
        for key_name in LAYER_KEYS:
            if key_name in ["ae13", "ab11"]:  # ABNT / JIS keys
                continue  # these two keys are not supported yet

            if key_name.startswith("-"):
                if output:
                    output.append("")
                output.append("<!--" + key_name[1:] + " -->")
                continue

            symbol = "&#x0010;"
            final_key = True

            if key_name in layer:
                key = layer[key_name]
                if key in layout.dead_keys:
                    symbol = f"dead_{DK_INDEX[key].name}"
                    final_key = False
                else:
                    symbol = xml_proof(key.upper() if caps else key)
                    final_key = not has_dead_keys(key.upper())

            char = f"code=\"{SCAN_CODES['osx'][key_name]}\"".ljust(10)
            if final_key:
                action = f'output="{symbol}"'
            else:
                action = f'action="{xml_proof_id(symbol)}"'
            output.append(f"<key {char} {action} />")

        ret_str.append(output)
    return ret_str


def osx_actions(layout: "KeyboardLayout") -> List[str]:
    """macOS layout, dead key actions."""

    ret_actions = []

    def when(state: str, action: str) -> str:
        state_attr = f'state="{state}"'.ljust(18)
        if action in layout.dead_keys:
            action_attr = f'next="{DK_INDEX[action].name}"'
        elif action.startswith("dead_"):
            action_attr = f'next="{action[5:]}"'
        else:
            action_attr = f'output="{xml_proof(action)}"'
        return f"  <when {state_attr} {action_attr} />"

    def append_actions(symbol: str, actions: List[Tuple[str, str]]) -> None:
        ret_actions.append(f'<action id="{xml_proof_id(symbol)}">')
        ret_actions.append(when("none", symbol))
        for state, out in actions:
            ret_actions.append(when(state, out))
        ret_actions.append("</action>")

    # dead key definitions
    for key in layout.dead_keys:
        name = DK_INDEX[key].name
        term = layout.dead_keys[key][key]
        ret_actions.append(f'<action id="dead_{name}">')
        ret_actions.append(f'  <when state="none" next="{name}" />')
        if name == "1dk" and term in layout.dead_keys:
            nested_dk = DK_INDEX[term].name
            ret_actions.append(f'  <when state="1dk" next="{nested_dk}" />')
        ret_actions.append("</action>")
        continue

    # normal key actions
    for key_name in LAYER_KEYS:
        if key_name.startswith("-"):
            ret_actions.append("")
            ret_actions.append(f"<!--{key_name[1:]} -->")
            continue

        for i in [Layer.BASE, Layer.SHIFT]:
            if key_name == "spce" or key_name not in layout.layers[i]:
                continue

            key = layout.layers[i][key_name]
            if i and key == layout.layers[Layer.BASE][key_name]:
                continue
            if key in layout.dead_keys:
                continue

            actions: List[Tuple[str, str]] = []
            for k in DK_INDEX:
                if k in layout.dead_keys:
                    if key in layout.dead_keys[k]:
                        actions.append((DK_INDEX[k].name, layout.dead_keys[k][key]))
            if actions:
                append_actions(xml_proof(key), actions)

    # spacebar actions
    actions = []
    for k in DK_INDEX:
        if k in layout.dead_keys:
            actions.append((DK_INDEX[k].name, layout.dead_keys[k][" "]))
    append_actions("&#x0020;", actions)  # space
    append_actions("&#x00a0;", actions)  # no-break space
    append_actions("&#x202f;", actions)  # fine no-break space

    return ret_actions


def osx_terminators(layout: "KeyboardLayout") -> List[str]:
    """macOS layout, dead key terminators."""

    ret_terminators = []
    for key in DK_INDEX:
        if key not in layout.dead_keys:
            continue
        dk = layout.dead_keys[key]
        name = DK_INDEX[key].name
        term = dk[key]
        if name == "1dk" and term in layout.dead_keys:
            term = dk[" "]
        state = f'state="{name}"'.ljust(18)
        output = f'output="{xml_proof(term)}"'
        ret_terminators.append(f"<when {state} {output} />")
    return ret_terminators


###
# Web: JSON
# To be used with the <x-keyboard> web component.
# https://github.com/OneDeadKey/x-keyboard
#


def web_keymap(layout: "KeyboardLayout") -> Dict[str, List[str]]:
    """Web layout, main part.

    Returns
    -------
    dict[str, list[str]]
        A dict whose keys are key ids and values are list of characters (length 2-4).
    """

    keymap = {}
    for key_name in LAYER_KEYS:
        if key_name.startswith("-"):
            continue
        chars = list("")
        for i in [Layer.BASE, Layer.SHIFT, Layer.ALTGR, Layer.ALTGR_SHIFT]:
            if key_name in layout.layers[i]:
                chars.append(layout.layers[i][key_name])
        if chars:
            keymap[SCAN_CODES["web"][key_name]] = chars

    return keymap


def web_deadkeys(layout: "KeyboardLayout") -> Dict[str, Dict[str, str]]:
    """Web layout, dead keys.

    Returns
    -------
    dict[str, dict[str, str]]
        A dict whose keys are deadkeys and values are key mapping.
    """

    return layout.dead_keys



import re
import sys
import traceback
from os import environ
from pathlib import Path
from textwrap import dedent
from typing import Dict, ItemsView, Optional

import click
from lxml import etree  # type: ignore
from lxml.builder import E  # type: ignore
from lxml.etree import _Element  # type: ignore

from .layout import KeyboardLayout


def xdg_config_home() -> Path:
    xdg_config = environ.get("XDG_CONFIG_HOME")
    if xdg_config:
        return Path(xdg_config)
    return Path.home() / ".config"


def wayland_running() -> bool:
    xdg_session = environ.get("XDG_SESSION_TYPE")
    if xdg_session:
        return xdg_session.startswith("wayland")
    return False


XKB_HOME = xdg_config_home() / "xkb"
XKB_ROOT = Path(environ.get("XKB_CONFIG_ROOT") or "/usr/share/X11/xkb/")

WAYLAND = wayland_running()

LayoutName = str
LocaleName = str
Variant = Dict[LayoutName, Optional[KeyboardLayout]]
Index = Dict[LocaleName, Variant]


class XKBManager:
    """Wrapper to list/add/remove keyboard drivers to XKB."""

    def __init__(self, root: bool = False) -> None:
        self._as_root = root
        self._rootdir = XKB_ROOT if root else XKB_HOME
        self._index: Index = {}

    @property
    def index(self) -> ItemsView[LocaleName, Variant]:
        return self._index.items()

    @property
    def path(self) -> Path:
        return self._rootdir

    def add(self, layout: KeyboardLayout) -> None:
        locale = layout.meta["locale"]
        variant = layout.meta["variant"]
        if locale not in self._index:
            self._index[locale] = {}
        self._index[locale][variant] = layout

    def remove(self, locale: str, variant: str) -> None:
        if locale not in self._index:
            self._index[locale] = {}
        self._index[locale][variant] = None

    def update(self) -> None:
        update_rules(self._rootdir, self._index)  # XKB/rules/{base,evdev}.xml
        update_symbols(self._rootdir, self._index)  # XKB/symbols/{locales}
        self._index = {}

    def clean(self) -> None:
        """Drop the obsolete 'type' attributes Kalamine used to add."""
        for filename in ["base.xml", "evdev.xml"]:
            filepath = self._rootdir / "rules" / filename
            if not filepath.exists():
                continue
            tree = etree.parse(str(filepath), etree.XMLParser(remove_blank_text=True))
            for variant in tree.xpath("//variant[@type]"):
                variant.attrib.pop("type")

    def list(self, mask: str = "") -> Index:
        layouts = list_rules(self._rootdir, mask)
        return list_symbols(self._rootdir, layouts)

    def list_all(self, mask: str = "") -> Index:
        return list_rules(self._rootdir, mask)

    def has_custom_symbols(self) -> bool:
        """Check if there is a usable xkb/symbols/custom file."""

        custom_path = self._rootdir / "symbols" / "custom"
        if not custom_path.exists():
            return False

        for filename in ["base.xml", "evdev.xml"]:
            filepath = self._rootdir / "rules" / filename
            if not filepath.exists():
                continue
            tree = etree.parse(str(filepath))
            if tree.xpath('//layout/configItem/name[text()="custom"]'):
                return True

        return False

    def ensure_xkb_config_is_ready(self) -> None:
        """Ensure there is an XKB configuration in user-space."""
        # See xkblayout.py for a more extensive version of this feature:
        # https://gitlab.freedesktop.org/whot/xkblayout

        if self._as_root:
            return

        # ensure all expected directories exist (don't care about 'geometry')
        XKB_HOME.mkdir(exist_ok=True)
        for subdir in ["compat", "keycodes", "rules", "symbols", "types"]:
            (XKB_HOME / subdir).mkdir(exist_ok=True)

        # ensure there are XKB rules
        # (new locales and symbols will be added by XKBManager)
        for ruleset in ["evdev"]:  # add 'base', too?
            # xkb/rules/evdev
            rules = XKB_HOME / "rules" / ruleset
            if not rules.exists():
                rules.write_text(
                    dedent(
                        f"""
                        // Generated by Kalamine
                        // Include the system '{ruleset}' file
                        ! include %S/{ruleset}
                        """
                    )
                )
            # xkb/rules/evdev.xml
            xmlpath = XKB_HOME / "rules" / f"{ruleset}.xml"
            if not xmlpath.exists():
                xmlpath.write_text(
                    dedent(
                        """\
                        <?xml version="1.0" encoding="UTF-8"?>
                        <!DOCTYPE xkbConfigRegistry SYSTEM "xkb.dtd">
                        <!-- Generated by Kalamine -->
                        <xkbConfigRegistry version="1.1">
                            <layoutList/>
                        </xkbConfigRegistry>
                        """
                    )
                )


""" On GNU/Linux, keyboard layouts must be installed in /usr/share/X11/xkb. To
    be able to revert a layout installation, Kalamine marks layouts like this:

    - XKB/symbols/[locale]: layout definitions
        // KALAMINE::[NAME]::BEGIN
        xkb_symbols "[name]" { ... }
        // KALAMINE::[NAME]::END

    Earlier versions of XKalamine used to mark index files as well but recent
    versions of Gnome do not support the custom `type` attribute any more, which
    must be removed:

    - XKB/rules/{base,evdev}.xml: layout references
        <variant type="kalamine">
            <configItem>
                <name>lafayette42</name>
                <description>French (Lafayette42)</description>
            </configItem>
        </variant>

    Even worse, the Lafayette project has released a first installer before
    the XKalamine installer was developed, so we have to handle this situation
    too:

    - XKB/symbols/[locale]: layout definitions
        // LAFAYETTE::BEGIN
        xkb_symbols "lafayette"   { ... }
        xkb_symbols "lafayette42" { ... }
        // LAFAYETTE::END

    - XKB/rules/{base,evdev}.xml: layout references
        <variant type="lafayette">
            <configItem>
                <name>lafayette</name>
                <description>French (Lafayette)</description>
            </configItem>
        </variant>
        <variant type="lafayette">
            <configItem>
                <name>lafayette42</name>
                <description>French (Lafayette42)</description>
            </configItem>
        </variant>

    Consequence: these two Lafayette layouts must be uninstalled together.
    Because of the way they are grouped in symbols/fr, it is impossible to
    remove one without removing the other.
"""


def clean_legacy_lafayette() -> None:
    return


###############################################################################
# Helpers: XKB/symbols
#


LEGACY_MARK = {"begin": "// LAFAYETTE::BEGIN\n", "end": "// LAFAYETTE::END\n"}


def get_symbol_mark(name: str) -> Dict[str, str]:
    return {
        "begin": "// KALAMINE::" + name.upper() + "::BEGIN\n",
        "end": "// KALAMINE::" + name.upper() + "::END\n",
    }


def is_new_symbol_mark(line: str) -> Optional[str]:
    if not line.endswith("::BEGIN\n"):
        return None

    if line.startswith("// KALAMINE::"):
        return line[13:-8].lower()  # XXX Kalamine expects lowercase names

    return "lafayette"  # line.startswith("// LAFAYETTE::"):  # obsolete marker


def update_symbols_locale(path: Path, named_layouts: Variant) -> None:
    """Update Kalamine layouts in an xkb/symbols/[locale] file."""

    text = ""
    modified_text = False
    with path.open("r+", encoding="utf-8") as symbols:
        # look for Kalamine layouts to be updated or removed
        between_marks = False
        closing_mark = ""
        for line in symbols:
            name = is_new_symbol_mark(line)
            if name:
                if name in named_layouts.keys():
                    closing_mark = line[:-6] + "END\n"
                    modified_text = True
                    between_marks = True
                    text = text.rstrip()
                else:
                    text += line
            elif line.endswith("::END\n"):
                if between_marks and line.startswith(closing_mark):
                    between_marks = False
                    closing_mark = ""
                else:
                    text += line
            elif not between_marks:
                text += line

        # clear previous Kalamine layouts if needed
        if modified_text:
            symbols.seek(0)
            symbols.write(text.rstrip() + "\n")
            symbols.truncate()

        # add new Kalamine layouts
        locale = path.name
        for name, layout in named_layouts.items():
            if layout is None:
                print(f"      - {locale}/{name}")
            else:
                print(f"      + {locale}/{name}")
                mark = get_symbol_mark(name)
                symbols.write("\n")
                symbols.write(mark["begin"])
                symbols.write(
                    re.sub(  # drop lines starting with '//#'
                        r"^//#.*\n", "", layout.xkb_symbols, flags=re.MULTILINE
                    ).rstrip()
                    + "\n"
                )
                symbols.write(mark["end"])

        symbols.close()


def update_symbols(xkb_root: Path, kb_index: Index) -> None:
    """Update Kalamine layouts in all xkb/symbols files."""

    for locale, named_layouts in kb_index.items():
        path = xkb_root / "symbols" / locale
        if not path.exists():
            with path.open("w") as file:
                file.write("// Generated by Kalamine")
                file.close()

        try:
            click.echo(f"... {path}")
            update_symbols_locale(path, named_layouts)

        except Exception as exc:
            exit_FileNotWritable(exc, path)


def list_symbols(xkb_root: Path, kb_index: Index) -> Index:
    """Filter input layouts: only keep the ones defined with Kalamine."""

    filtered_index: Index = {}
    for locale, variants in sorted(kb_index.items()):
        path = xkb_root / "symbols" / locale
        if not path.exists():
            continue

        with open(path, "r", encoding="utf-8") as symbols:
            for line in symbols:
                name = is_new_symbol_mark(line)
                if name is None:
                    continue
                if name in variants.keys():
                    if locale not in filtered_index:
                        filtered_index[locale] = {}
                    filtered_index[locale][name] = variants[name]

    return filtered_index


###############################################################################
# Helpers: XKB/rules
#


def get_rules_locale(tree: etree.ElementTree, locale: LocaleName) -> _Element:
    query = f'//layout/configItem/name[text()="{locale}"]/../..'
    result = tree.xpath(query)
    if len(result) != 1:
        tree.xpath("//layoutList")[0].append(
            E.layout(E.configItem(E.name(locale)), E.variantList())
        )
    return tree.xpath(query)[0]


def remove_rules_variant(variant_list: _Element, name: str) -> None:
    query = f'variant/configItem/name[text()="{name}"]/../..'
    for variant in variant_list.xpath(query):
        variant.getparent().remove(variant)


def add_rules_variant(variant_list: _Element, name: str, description: str) -> None:
    variant_list.append(
        E.variant(E.configItem(E.name(name), E.description(description)))
    )


def update_rules(xkb_root: Path, kb_index: Index) -> None:
    """Update references in XKB/rules/{base,evdev}.xml."""

    for filename in ["base.xml", "evdev.xml"]:
        filepath = xkb_root / "rules" / filename
        if not filepath.exists():
            continue

        try:
            tree = etree.parse(filepath, etree.XMLParser(remove_blank_text=True))

            for locale, named_layouts in kb_index.items():
                vlist = get_rules_locale(tree, locale).xpath("variantList")
                if len(vlist) != 1:
                    exit(f"Error: unexpected xml format in {filepath}.")
                for name, layout in named_layouts.items():
                    remove_rules_variant(vlist[0], name)
                    if layout is not None:
                        description = layout.meta["description"]
                        add_rules_variant(vlist[0], name, description)

            tree.write(
                filepath, pretty_print=True, xml_declaration=True, encoding="utf-8"
            )
            click.echo(f"... {filepath}")

        except Exception as exc:
            exit_FileNotWritable(exc, filepath)


def list_rules(xkb_root: Path, mask: str = "*") -> Index:
    """List all matching XKB layouts."""

    if mask in ("", "*"):
        locale_mask = "*"
        variant_mask = "*"
    else:
        m = mask.split("/")
        if len(m) == 2:
            locale_mask, variant_mask = m
        else:
            locale_mask = mask
            variant_mask = "*"

    kb_index: Index = {}
    for filename in ["base.xml", "evdev.xml"]:
        filepath = xkb_root / "rules" / filename
        if not filepath.exists():
            continue

        tree = etree.parse(filepath)
        for variant in tree.xpath("//variant"):
            locale = variant.xpath("../../configItem/name")[0].text
            name = variant.xpath("configItem/name")[0].text
            desc = variant.xpath("configItem/description")[0].text

            if locale_mask in ("*", locale) and variant_mask in ("*", name):
                if locale not in kb_index:
                    kb_index[locale] = {}
                kb_index[locale][name] = desc

    return kb_index


###############################################################################
# Exception Handling (there must be a better way...)
#


def exit_FileNotWritable(exception: Exception, path: Path) -> None:
    if isinstance(exception, PermissionError):  # noqa: F821
        raise exception
    if isinstance(exception, IOError):
        click.echo("")
        sys.exit(f"Error: could not write to file {path}.")
    else:
        click.echo("")
        sys.exit(f"Error: {exception}.\n{traceback.format_exc()}")



from textwrap import dedent

from kalamine import KeyboardLayout
from kalamine.template import xkb_keymap

from .util import get_layout_dict


def load_layout(filename: str) -> KeyboardLayout:
    return KeyboardLayout(get_layout_dict(filename))


def split(multiline_str: str):
    return dedent(multiline_str).lstrip().rstrip().splitlines()


def test_ansi():
    layout = load_layout("ansi")

    expected = split(
        """
        // Digits
        key <AE01> {[ 1               , exclam          , VoidSymbol      , VoidSymbol      ]}; // 1 !
        key <AE02> {[ 2               , at              , VoidSymbol      , VoidSymbol      ]}; // 2 @
        key <AE03> {[ 3               , numbersign      , VoidSymbol      , VoidSymbol      ]}; // 3 #
        key <AE04> {[ 4               , dollar          , VoidSymbol      , VoidSymbol      ]}; // 4 $
        key <AE05> {[ 5               , percent         , VoidSymbol      , VoidSymbol      ]}; // 5 %
        key <AE06> {[ 6               , asciicircum     , VoidSymbol      , VoidSymbol      ]}; // 6 ^
        key <AE07> {[ 7               , ampersand       , VoidSymbol      , VoidSymbol      ]}; // 7 &
        key <AE08> {[ 8               , asterisk        , VoidSymbol      , VoidSymbol      ]}; // 8 *
        key <AE09> {[ 9               , parenleft       , VoidSymbol      , VoidSymbol      ]}; // 9 (
        key <AE10> {[ 0               , parenright      , VoidSymbol      , VoidSymbol      ]}; // 0 )

        // Letters, first row
        key <AD01> {[ q               , Q               , VoidSymbol      , VoidSymbol      ]}; // q Q
        key <AD02> {[ w               , W               , VoidSymbol      , VoidSymbol      ]}; // w W
        key <AD03> {[ e               , E               , VoidSymbol      , VoidSymbol      ]}; // e E
        key <AD04> {[ r               , R               , VoidSymbol      , VoidSymbol      ]}; // r R
        key <AD05> {[ t               , T               , VoidSymbol      , VoidSymbol      ]}; // t T
        key <AD06> {[ y               , Y               , VoidSymbol      , VoidSymbol      ]}; // y Y
        key <AD07> {[ u               , U               , VoidSymbol      , VoidSymbol      ]}; // u U
        key <AD08> {[ i               , I               , VoidSymbol      , VoidSymbol      ]}; // i I
        key <AD09> {[ o               , O               , VoidSymbol      , VoidSymbol      ]}; // o O
        key <AD10> {[ p               , P               , VoidSymbol      , VoidSymbol      ]}; // p P

        // Letters, second row
        key <AC01> {[ a               , A               , VoidSymbol      , VoidSymbol      ]}; // a A
        key <AC02> {[ s               , S               , VoidSymbol      , VoidSymbol      ]}; // s S
        key <AC03> {[ d               , D               , VoidSymbol      , VoidSymbol      ]}; // d D
        key <AC04> {[ f               , F               , VoidSymbol      , VoidSymbol      ]}; // f F
        key <AC05> {[ g               , G               , VoidSymbol      , VoidSymbol      ]}; // g G
        key <AC06> {[ h               , H               , VoidSymbol      , VoidSymbol      ]}; // h H
        key <AC07> {[ j               , J               , VoidSymbol      , VoidSymbol      ]}; // j J
        key <AC08> {[ k               , K               , VoidSymbol      , VoidSymbol      ]}; // k K
        key <AC09> {[ l               , L               , VoidSymbol      , VoidSymbol      ]}; // l L
        key <AC10> {[ semicolon       , colon           , VoidSymbol      , VoidSymbol      ]}; // ; :

        // Letters, third row
        key <AB01> {[ z               , Z               , VoidSymbol      , VoidSymbol      ]}; // z Z
        key <AB02> {[ x               , X               , VoidSymbol      , VoidSymbol      ]}; // x X
        key <AB03> {[ c               , C               , VoidSymbol      , VoidSymbol      ]}; // c C
        key <AB04> {[ v               , V               , VoidSymbol      , VoidSymbol      ]}; // v V
        key <AB05> {[ b               , B               , VoidSymbol      , VoidSymbol      ]}; // b B
        key <AB06> {[ n               , N               , VoidSymbol      , VoidSymbol      ]}; // n N
        key <AB07> {[ m               , M               , VoidSymbol      , VoidSymbol      ]}; // m M
        key <AB08> {[ comma           , less            , VoidSymbol      , VoidSymbol      ]}; // , <
        key <AB09> {[ period          , greater         , VoidSymbol      , VoidSymbol      ]}; // . >
        key <AB10> {[ slash           , question        , VoidSymbol      , VoidSymbol      ]}; // / ?

        // Pinky keys
        key <AE11> {[ minus           , underscore      , VoidSymbol      , VoidSymbol      ]}; // - _
        key <AE12> {[ equal           , plus            , VoidSymbol      , VoidSymbol      ]}; // = +
        key <AE13> {[ VoidSymbol      , VoidSymbol      , VoidSymbol      , VoidSymbol      ]}; //
        key <AD11> {[ bracketleft     , braceleft       , VoidSymbol      , VoidSymbol      ]}; // [ {
        key <AD12> {[ bracketright    , braceright      , VoidSymbol      , VoidSymbol      ]}; // ] }
        key <AC11> {[ apostrophe      , quotedbl        , VoidSymbol      , VoidSymbol      ]}; // ' "
        key <AB11> {[ VoidSymbol      , VoidSymbol      , VoidSymbol      , VoidSymbol      ]}; //
        key <TLDE> {[ grave           , asciitilde      , VoidSymbol      , VoidSymbol      ]}; // ` ~
        key <BKSL> {[ backslash       , bar             , VoidSymbol      , VoidSymbol      ]}; // \\ |
        key <LSGT> {[ VoidSymbol      , VoidSymbol      , VoidSymbol      , VoidSymbol      ]}; //

        // Space bar
        key <SPCE> {[ space           , space           , apostrophe      , apostrophe      ]}; //     ' '
        """
    )

    xkbcomp = xkb_keymap(layout, xkbcomp=True)
    assert len(xkbcomp) == len(expected)
    assert xkbcomp == expected

    xkbpatch = xkb_keymap(layout, xkbcomp=False)
    assert len(xkbpatch) == len(expected)
    assert xkbpatch == expected


def test_intl():
    layout = load_layout("intl")

    expected = split(
        """
        // Digits
        key <AE01> {[ 1               , exclam          , VoidSymbol      , VoidSymbol      ]}; // 1 !
        key <AE02> {[ 2               , at              , VoidSymbol      , VoidSymbol      ]}; // 2 @
        key <AE03> {[ 3               , numbersign      , VoidSymbol      , VoidSymbol      ]}; // 3 #
        key <AE04> {[ 4               , dollar          , VoidSymbol      , VoidSymbol      ]}; // 4 $
        key <AE05> {[ 5               , percent         , VoidSymbol      , VoidSymbol      ]}; // 5 %
        key <AE06> {[ 6               , dead_circumflex , VoidSymbol      , VoidSymbol      ]}; // 6 ^
        key <AE07> {[ 7               , ampersand       , VoidSymbol      , VoidSymbol      ]}; // 7 &
        key <AE08> {[ 8               , asterisk        , VoidSymbol      , VoidSymbol      ]}; // 8 *
        key <AE09> {[ 9               , parenleft       , VoidSymbol      , VoidSymbol      ]}; // 9 (
        key <AE10> {[ 0               , parenright      , VoidSymbol      , VoidSymbol      ]}; // 0 )

        // Letters, first row
        key <AD01> {[ q               , Q               , VoidSymbol      , VoidSymbol      ]}; // q Q
        key <AD02> {[ w               , W               , VoidSymbol      , VoidSymbol      ]}; // w W
        key <AD03> {[ e               , E               , eacute          , Eacute          ]}; // e E é É
        key <AD04> {[ r               , R               , VoidSymbol      , VoidSymbol      ]}; // r R
        key <AD05> {[ t               , T               , VoidSymbol      , VoidSymbol      ]}; // t T
        key <AD06> {[ y               , Y               , VoidSymbol      , VoidSymbol      ]}; // y Y
        key <AD07> {[ u               , U               , uacute          , Uacute          ]}; // u U ú Ú
        key <AD08> {[ i               , I               , iacute          , Iacute          ]}; // i I í Í
        key <AD09> {[ o               , O               , oacute          , Oacute          ]}; // o O ó Ó
        key <AD10> {[ p               , P               , VoidSymbol      , VoidSymbol      ]}; // p P

        // Letters, second row
        key <AC01> {[ a               , A               , aacute          , Aacute          ]}; // a A á Á
        key <AC02> {[ s               , S               , VoidSymbol      , VoidSymbol      ]}; // s S
        key <AC03> {[ d               , D               , VoidSymbol      , VoidSymbol      ]}; // d D
        key <AC04> {[ f               , F               , VoidSymbol      , VoidSymbol      ]}; // f F
        key <AC05> {[ g               , G               , VoidSymbol      , VoidSymbol      ]}; // g G
        key <AC06> {[ h               , H               , VoidSymbol      , VoidSymbol      ]}; // h H
        key <AC07> {[ j               , J               , VoidSymbol      , VoidSymbol      ]}; // j J
        key <AC08> {[ k               , K               , VoidSymbol      , VoidSymbol      ]}; // k K
        key <AC09> {[ l               , L               , VoidSymbol      , VoidSymbol      ]}; // l L
        key <AC10> {[ semicolon       , colon           , VoidSymbol      , VoidSymbol      ]}; // ; :

        // Letters, third row
        key <AB01> {[ z               , Z               , VoidSymbol      , VoidSymbol      ]}; // z Z
        key <AB02> {[ x               , X               , VoidSymbol      , VoidSymbol      ]}; // x X
        key <AB03> {[ c               , C               , ccedilla        , Ccedilla        ]}; // c C ç Ç
        key <AB04> {[ v               , V               , VoidSymbol      , VoidSymbol      ]}; // v V
        key <AB05> {[ b               , B               , VoidSymbol      , VoidSymbol      ]}; // b B
        key <AB06> {[ n               , N               , VoidSymbol      , VoidSymbol      ]}; // n N
        key <AB07> {[ m               , M               , VoidSymbol      , VoidSymbol      ]}; // m M
        key <AB08> {[ comma           , less            , VoidSymbol      , VoidSymbol      ]}; // , <
        key <AB09> {[ period          , greater         , ellipsis        , VoidSymbol      ]}; // . > …
        key <AB10> {[ slash           , question        , VoidSymbol      , VoidSymbol      ]}; // / ?

        // Pinky keys
        key <AE11> {[ minus           , underscore      , VoidSymbol      , VoidSymbol      ]}; // - _
        key <AE12> {[ equal           , plus            , VoidSymbol      , VoidSymbol      ]}; // = +
        key <AE13> {[ VoidSymbol      , VoidSymbol      , VoidSymbol      , VoidSymbol      ]}; //
        key <AD11> {[ bracketleft     , braceleft       , VoidSymbol      , VoidSymbol      ]}; // [ {
        key <AD12> {[ bracketright    , braceright      , VoidSymbol      , VoidSymbol      ]}; // ] }
        key <AC11> {[ ISO_Level3_Latch, dead_diaeresis  , acute           , VoidSymbol      ]}; // ´ ¨ ´
        key <AB11> {[ VoidSymbol      , VoidSymbol      , VoidSymbol      , VoidSymbol      ]}; //
        key <TLDE> {[ dead_grave      , dead_tilde      , VoidSymbol      , VoidSymbol      ]}; // ` ~
        key <BKSL> {[ backslash       , bar             , VoidSymbol      , VoidSymbol      ]}; // \\ |
        key <LSGT> {[ backslash       , bar             , VoidSymbol      , VoidSymbol      ]}; // \\ |

        // Space bar
        key <SPCE> {[ space           , space           , apostrophe      , apostrophe      ]}; //     ' '
        """
    )

    xkbcomp = xkb_keymap(layout, xkbcomp=True)
    assert len(xkbcomp) == len(expected)
    assert xkbcomp == expected

    xkbpatch = xkb_keymap(layout, xkbcomp=False)
    assert len(xkbpatch) == len(expected)
    assert xkbpatch == expected


def test_prog():
    layout = load_layout("prog")

    expected = split(
        """
        // Digits
        key <AE01> {[ 1               , exclam          , exclam          , VoidSymbol      ]}; // 1 ! !
        key <AE02> {[ 2               , at              , parenleft       , VoidSymbol      ]}; // 2 @ (
        key <AE03> {[ 3               , numbersign      , parenright      , VoidSymbol      ]}; // 3 # )
        key <AE04> {[ 4               , dollar          , apostrophe      , VoidSymbol      ]}; // 4 $ '
        key <AE05> {[ 5               , percent         , quotedbl        , VoidSymbol      ]}; // 5 % "
        key <AE06> {[ 6               , asciicircum     , dead_circumflex , VoidSymbol      ]}; // 6 ^ ^
        key <AE07> {[ 7               , ampersand       , 7               , VoidSymbol      ]}; // 7 & 7
        key <AE08> {[ 8               , asterisk        , 8               , VoidSymbol      ]}; // 8 * 8
        key <AE09> {[ 9               , parenleft       , 9               , VoidSymbol      ]}; // 9 ( 9
        key <AE10> {[ 0               , parenright      , slash           , VoidSymbol      ]}; // 0 ) /

        // Letters, first row
        key <AD01> {[ q               , Q               , equal           , VoidSymbol      ]}; // q Q =
        key <AD02> {[ w               , W               , less            , lessthanequal   ]}; // w W < ≤
        key <AD03> {[ e               , E               , greater         , greaterthanequal]}; // e E > ≥
        key <AD04> {[ r               , R               , minus           , VoidSymbol      ]}; // r R -
        key <AD05> {[ t               , T               , plus            , VoidSymbol      ]}; // t T +
        key <AD06> {[ y               , Y               , VoidSymbol      , VoidSymbol      ]}; // y Y
        key <AD07> {[ u               , U               , 4               , VoidSymbol      ]}; // u U 4
        key <AD08> {[ i               , I               , 5               , VoidSymbol      ]}; // i I 5
        key <AD09> {[ o               , O               , 6               , VoidSymbol      ]}; // o O 6
        key <AD10> {[ p               , P               , asterisk        , VoidSymbol      ]}; // p P *

        // Letters, second row
        key <AC01> {[ a               , A               , braceleft       , VoidSymbol      ]}; // a A {
        key <AC02> {[ s               , S               , bracketleft     , VoidSymbol      ]}; // s S [
        key <AC03> {[ d               , D               , bracketright    , VoidSymbol      ]}; // d D ]
        key <AC04> {[ f               , F               , braceright      , VoidSymbol      ]}; // f F }
        key <AC05> {[ g               , G               , slash           , VoidSymbol      ]}; // g G /
        key <AC06> {[ h               , H               , VoidSymbol      , VoidSymbol      ]}; // h H
        key <AC07> {[ j               , J               , 1               , VoidSymbol      ]}; // j J 1
        key <AC08> {[ k               , K               , 2               , VoidSymbol      ]}; // k K 2
        key <AC09> {[ l               , L               , 3               , VoidSymbol      ]}; // l L 3
        key <AC10> {[ semicolon       , colon           , minus           , VoidSymbol      ]}; // ; : -

        // Letters, third row
        key <AB01> {[ z               , Z               , asciitilde      , VoidSymbol      ]}; // z Z ~
        key <AB02> {[ x               , X               , grave           , VoidSymbol      ]}; // x X `
        key <AB03> {[ c               , C               , bar             , brokenbar       ]}; // c C | ¦
        key <AB04> {[ v               , V               , underscore      , VoidSymbol      ]}; // v V _
        key <AB05> {[ b               , B               , backslash       , VoidSymbol      ]}; // b B \\ 
        key <AB06> {[ n               , N               , VoidSymbol      , VoidSymbol      ]}; // n N
        key <AB07> {[ m               , M               , 0               , VoidSymbol      ]}; // m M 0
        key <AB08> {[ comma           , less            , comma           , VoidSymbol      ]}; // , < ,
        key <AB09> {[ period          , greater         , period          , VoidSymbol      ]}; // . > .
        key <AB10> {[ slash           , question        , plus            , VoidSymbol      ]}; // / ? +

        // Pinky keys
        key <AE11> {[ minus           , underscore      , VoidSymbol      , VoidSymbol      ]}; // - _
        key <AE12> {[ equal           , plus            , VoidSymbol      , VoidSymbol      ]}; // = +
        key <AE13> {[ VoidSymbol      , VoidSymbol      , VoidSymbol      , VoidSymbol      ]}; //
        key <AD11> {[ bracketleft     , braceleft       , VoidSymbol      , VoidSymbol      ]}; // [ {
        key <AD12> {[ bracketright    , braceright      , VoidSymbol      , VoidSymbol      ]}; // ] }
        key <AC11> {[ apostrophe      , quotedbl        , dead_acute      , dead_diaeresis  ]}; // ' " ´ ¨
        key <AB11> {[ VoidSymbol      , VoidSymbol      , VoidSymbol      , VoidSymbol      ]}; //
        key <TLDE> {[ grave           , asciitilde      , dead_grave      , dead_tilde      ]}; // ` ~ ` ~
        key <BKSL> {[ backslash       , bar             , VoidSymbol      , VoidSymbol      ]}; // \\ |
        key <LSGT> {[ VoidSymbol      , VoidSymbol      , VoidSymbol      , VoidSymbol      ]}; //

        // Space bar
        key <SPCE> {[ space           , space           , space           , space           ]}; //
        """
    )

    xkbcomp = xkb_keymap(layout, xkbcomp=True)
    assert len(xkbcomp) == len(expected)
    assert xkbcomp == expected

    xkbpatch = xkb_keymap(layout, xkbcomp=False)
    assert len(xkbpatch) == len(expected)
    assert xkbpatch == expected



import json
import threading
import webbrowser
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pathlib import Path

import click
from livereload import Server # type: ignore

from .layout import KeyboardLayout, load_layout


def keyboard_server(file_path: Path) -> None:
    kb_layout = KeyboardLayout(load_layout(file_path))

    host_name = "localhost"
    webserver_port = 1664
    lr_server_port = 5500

    def main_page(layout: KeyboardLayout) -> str:
        return f"""
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml">
            <head>
                <meta charset="utf-8" />
                <title>Kalamine</title>
                <link rel="stylesheet" type="text/css" href="style.css" />
                <script src="x-keyboard.js" type="module"></script>
                <script src="http://{host_name}:{lr_server_port}/livereload.js"></script>
                <script src="demo.js" type="text/javascript"></script>
            </head>
            <body>
                <p>
                    <a href="{layout.meta['url']}">{layout.meta['name']}</a>
                    <br /> {layout.meta['locale']}/{layout.meta['variant']}
                    <br /> {layout.meta['description']}
                </p>
                <input spellcheck="false" placeholder="" />
                <x-keyboard src="/json"></x-keyboard>
                <p style="text-align: right;">
                    <select>
                        <option value="iso">  ISO  </option>
                        <option value="ansi"> ANSI </option>
                        <option value="ol60"> ERGO </option>
                        <option value="ol50"> 4×12 </option>
                        <option value="ol40"> 3×12 </option>
                    </select>
                </p>
                <p style="text-align: center;">
                    <a href="/json">json</a>
                    | <a href="/keylayout">keylayout</a>
                    | <a href="/klc">klc</a>
                    | <a href="/xkb_keymap">xkb_keymap</a>
                    | <a href="/xkb_symbols">xkb_symbols</a>
                </p>
            </body>
            </html>
        """

    class LayoutHandler(SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs) -> None:  # type: ignore
            kwargs["directory"] = str(Path(__file__).parent / "www")
            super().__init__(*args, **kwargs)

        def do_GET(self) -> None:
            self.send_response(200)

            def send(
                page: str, content: str = "text/plain", charset: str = "utf-8"
            ) -> None:
                self.send_header("Content-type", f"{content}; charset={charset}")
                self.end_headers()
                self.wfile.write(bytes(page, charset))
                # self.wfile.write(page.encode(charset))

            # XXX always reloads the layout on the root page, never in sub pages
            nonlocal kb_layout
            if self.path == "/favicon.ico":
                pass
            elif self.path == "/json":
                send(json.dumps(kb_layout.json), content="application/json")
            elif self.path == "/keylayout":
                # send(kb_layout.keylayout, content='application/xml')
                send(kb_layout.keylayout)
            elif self.path == "/klc":
                send(kb_layout.klc, charset="utf-16-le", content="text")
            elif self.path == "/xkb_keymap":
                send(kb_layout.xkb_keymap)
            elif self.path == "/xkb_symbols":
                send(kb_layout.xkb_symbols.replace("//#", "//"))
            elif self.path == "/":
                kb_layout = KeyboardLayout(load_layout(file_path))  # refresh
                send(main_page(kb_layout), content="text/html")
            else:
                return SimpleHTTPRequestHandler.do_GET(self)

    webserver = HTTPServer((host_name, webserver_port), LayoutHandler)
    thread = threading.Thread(None, webserver.serve_forever)

    try:
        thread.start()
        url = f"http://{host_name}:{webserver_port}"
        print(f"Server started: {url}")
        print("Hit Ctrl-C to stop.")
        webbrowser.open(url)

        # livereload
        lr_server = Server()
        lr_server.watch(str(file_path))
        lr_server.serve(host=host_name, port=lr_server_port)

    except KeyboardInterrupt:
        pass

    webserver.shutdown()
    webserver.server_close()
    thread.join()
    click.echo("Server stopped.")



"""Some util functions for tests."""

import pkgutil
from typing import Dict

import tomli


def get_layout_dict(filename: str) -> Dict:
    """Return the layout directory path."""

    descriptor = pkgutil.get_data(__package__, f"../layouts/{filename}.toml")
    return tomli.loads(descriptor.decode("utf-8"))
